"""
Some functionality is based heavily on Flask-Marshmallow:
https://github.com/marshmallow-code/flask-marshmallow/blob/dev/src/flask_marshmallow/fields.py
"""

import abc
import re
import urllib
from typing import Any, Dict, List, Optional, Type, no_type_check

from fastapi import FastAPI
from pydantic import BaseModel, PrivateAttr, root_validator
from pydantic.utils import update_not_none
from pydantic.validators import dict_validator
from starlette.datastructures import URLPath
from starlette.routing import Route

_tpl_pattern = re.compile(r"\s*<\s*(\S*)\s*>\s*")

_uri_schema = {"type": "string", "format": "uri", "minLength": 1, "maxLength": 2 ** 16}


class InvalidAttribute(AttributeError):
    pass


class AbstractHyperField(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __build_hypermedia__(self, app: Optional[FastAPI], values: Dict[str, Any]):
        return


class UrlType(str):
    @classmethod
    def __modify_schema__(cls, field_schema: Dict[str, Any]) -> None:
        update_not_none(
            field_schema,
            **_uri_schema,
        )


class UrlFor(UrlType, AbstractHyperField):
    def __init__(self, endpoint: str, param_values: Optional[Dict[str, str]] = None):
        self.endpoint: str = endpoint
        self.param_values: Dict[str, str] = param_values or {}
        super().__init__()

    @no_type_check
    def __new__(cls, *_):
        return str.__new__(cls)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value: Any) -> "UrlFor":
        """
        Validate UrlFor object against itself.
        The UrlFor field type will only accept UrlFor instances.
        """
        # Return original object if it's already a UrlFor instance
        if value.__class__ == URLPath:
            return value
        # Otherwise raise an exception
        raise ValueError(
            f"UrlFor field should resolve to a starlette.datastructures.URLPath instance. Instead got {value.__class__}"
        )

    def __build_hypermedia__(
        self, app: Optional[FastAPI], values: Dict[str, Any]
    ) -> Optional[str]:
        if app is None:
            return None
        resolved_params = _resolve_param_values(self.param_values, values)
        return app.url_path_for(self.endpoint, **resolved_params)


class HALItem(BaseModel):
    href: Optional[UrlType]
    method: Optional[str]
    description: Optional[str]


class HALFor(HALItem, AbstractHyperField):
    _endpoint: str = PrivateAttr()
    _param_values: Optional[Dict[str, str]] = PrivateAttr()
    _description: Optional[str] = PrivateAttr()

    def __init__(
        self,
        endpoint: str,
        param_values: Optional[Dict[str, str]] = None,
        description: Optional[str] = None,
    ):
        self._endpoint: str = endpoint
        self._param_values: Dict[str, str] = param_values or {}
        self._description = description
        super().__init__()

    def __build_hypermedia__(
        self, app: Optional[FastAPI], values: Dict[str, Any]
    ) -> Optional[HALItem]:
        if app is None:
            return None
        resolved_params = _resolve_param_values(self._param_values, values)

        this_route = next(
            (
                route
                for route in app.routes
                if isinstance(route, Route) and route.name == self._endpoint
            ),
            None,
        )
        if not this_route:
            raise ValueError(f"No route found for endpoint {self._endpoint}")

        return HALItem(
            href=app.url_path_for(self._endpoint, **resolved_params),
            method=next(iter(this_route.methods), None) if this_route.methods else None,
            description=self._description,
        )


_LinkSetType = Dict[str, UrlFor]


class LinkSet(_LinkSetType, AbstractHyperField):  # pylint: disable=too-many-ancestors
    @classmethod
    def __get_validators__(cls):
        yield dict_validator

    @classmethod
    def __modify_schema__(cls, field_schema: Dict[str, Any]) -> None:
        field_schema.update({"additionalProperties": _uri_schema})

    def __build_hypermedia__(
        self, app: Optional[FastAPI], values: Dict[str, Any]
    ) -> Dict[str, str]:
        return {k: u.__build_hypermedia__(app, values) for k, u in self.items()}  # type: ignore  # pylint: disable=no-member


def _tpl(val):
    """Return value within ``< >`` if possible, else return ``None``."""
    match = _tpl_pattern.match(val)
    if match:
        return match.groups()[0]
    return None


def _get_value(obj: Any, key: str, default: Optional[Any] = None):
    """Slightly-modified version of marshmallow.utils.get_value.
    If a dot-delimited ``key`` is passed and any attribute in the
    path is `None`, return `None`.
    """
    if "." in key:
        return _get_value_for_keys(obj, key.split("."), default)
    return _get_value_for_key(obj, key, default)


def _get_value_for_keys(obj: Any, keys: List[str], default: Any):
    if len(keys) == 1:
        return _get_value_for_key(obj, keys[0], default)
    value = _get_value_for_key(obj, keys[0], default)
    if value is None:
        return None
    return _get_value_for_keys(value, keys[1:], default)


def _get_value_for_key(obj: Any, key: str, default: Any):
    if not hasattr(obj, "__getitem__"):
        return getattr(obj, key, default)

    try:
        return obj[key]
    except (KeyError, IndexError, TypeError, AttributeError):
        return getattr(obj, key, default)


def _clean_attribute_value(value: Any) -> str:
    if isinstance(value, str):
        return urllib.parse.quote(value)
    return value


def _resolve_param_values(param_values_template, data_object) -> Dict[str, str]:
    param_values = {}
    for name, attr_tpl in param_values_template.items():
        attr_name = _tpl(str(attr_tpl))
        if attr_name:
            attribute_value = _get_value(data_object, attr_name, default=None)
            if attribute_value is None:
                raise InvalidAttribute(
                    f"{attr_name} is not a valid attribute of {data_object}"
                )
            param_values[name] = _clean_attribute_value(attribute_value)
    return param_values


class HyperModel(BaseModel):
    _hypermodel_bound_app: Optional[FastAPI] = None

    # List of types with __build_hypermedia__ methods that we can process
    # Currently static but could enable extensions in future
    _hypermodel_bound_field_types: List[Type[AbstractHyperField]] = [
        UrlFor,
        LinkSet,
        HALFor,
    ]

    @classmethod
    def _generate_url(cls, endpoint, param_values, values):
        if cls._hypermodel_bound_app:
            return cls._hypermodel_bound_app.url_path_for(
                endpoint, **_resolve_param_values(param_values, values)
            )
        return None

    @root_validator
    def _hypermodel_gen_href(cls, values):  # pylint: disable=no-self-argument
        for key, value in values.items():
            if isinstance(value, tuple(cls._hypermodel_bound_field_types)):
                values[key] = value.__build_hypermedia__(
                    cls._hypermodel_bound_app, values
                )
        return values

    @classmethod
    def init_app(cls, app: FastAPI):
        """
        Bind a FastAPI app to othe HyperModel base class.
        This allows HyperModel to convert endpoint function names into
        working URLs relative to the application root.

        Args:
            app (FastAPI): Application to generate URLs from
        """
        cls._hypermodel_bound_app = app



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fastapi_hypermodel']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0.65.2', 'pydantic>=1.8.0,<2.0.0']

setup_kwargs = {
    'name': 'fastapi-hypermodel',
    'version': '0.3.2',
    'description': 'A FastAPI + Pydantic extension for simplifying hypermedia-driven API development.',
    'long_description': '# FastAPI-HyperModel\n\nFastAPI-HyperModel is a FastAPI + Pydantic extension for simplifying hypermedia-driven API development. This module adds a new Pydantic model base-class, supporting dynamic `href` generation based on object data.\n\n<table>\n<tbody>\n<tr>\n<th>Model</th>\n<th>Response</th>\n</tr>\n<tr>\n<td>\n\n```python\nclass ItemSummary(HyperModel):\n    name: str\n    id: str\n    href = UrlFor(\n        "read_item", {"item_id": "<id>"}\n    )\n```\n\n</td>\n<td>\n\n```json\n{\n  "name": "Foo",\n  "id": "item01",\n  "href": "/items/item01"\n}\n```\n\n</td>\n</tr>\n<tr></tr>\n<tr>\n<td>\n\n```python\nclass ItemSummary(HyperModel):\n    name: str\n    id: str\n    link = HALFor(\n        "read_item", {"item_id": "<id>"}, \n        description="Read an item"\n    )\n```\n\n</td>\n<td>\n\n```json\n{\n  "name": "Foo",\n  "id": "item01",\n  "link": {\n      "href": "/items/item01",\n      "method": "GET",\n      "description": "Read an item"\n  }\n}\n```\n\n</td>\n</tr>\n</tbody>\n</table>\n\n## Installation\n\n`pip install fastapi-hypermodel`\n\n## Basic Usage\n\n### Import `HyperModel` and optionally `HyperRef`\n\n```python\nfrom fastapi import FastAPI\n\nfrom fastapi_hypermodel import HyperModel, UrlFor, LinkSet\n```\n\n`HyperModel` will be your model base-class.\n\n### Create your basic models\n\nWe\'ll create two models, a brief item summary including ID, name, and a link, and a full model containing additional information. We\'ll use `ItemSummary` in our item list, and `ItemDetail` for full item information.\n\n```python\nclass ItemSummary(HyperModel):\n    id: str\n    name: str\n\nclass ItemDetail(ItemSummary):\n    description: Optional[str] = None\n    price: float\n\nclass Person(HyperModel):\n    name: str\n    id: str\n    items: List[ItemSummary]\n```\n\n### Create and attach your app\n\nWe\'ll now create our FastAPI app, and bind it to our `HyperModel` base class.\n\n```python\nfrom fastapi import FastAPI\n\napp = FastAPI()\nHyperModel.init_app(app)\n```\n\n### Add some API views\n\nWe\'ll create an API view for a list of items, as well as details about an individual item. Note that we pass the item ID with our `{item_id}` URL variable.\n\n```python\n@app.get("/items", response_model=List[ItemSummary])\ndef read_items():\n    return list(items.values())\n\n@app.get("/items/{item_id}", response_model=ItemDetail)\ndef read_item(item_id: str):\n    return items[item_id]\n\n@app.get("/people/{person_id}", response_model=Person)\ndef read_person(person_id: str):\n    return people[person_id]\n\n@app.get("/people/{person_id}/items", response_model=List[ItemDetail])\ndef read_person_items(person_id: str):\n    return people[person_id]["items"]\n```\n\n### Create a model `href`\n\nWe\'ll now go back and add an `href` field with a special `UrlFor` value. This `UrlFor` class defines how our href elements will be generated. We\'ll change our `ItemSummary` class to:\n\n```python\nclass ItemSummary(HyperModel):\n    name: str\n    id: str\n    href = UrlFor("read_item", {"item_id": "<id>"})\n```\n\nThe `UrlFor` class takes two arguments:\n\n#### `endpoint`\n\nName of your FastAPI endpoint function you want to link to. In our example, we want our item summary to link to the corresponding item detail page, which maps to our `read_item` function.\n\n#### `values` (optional depending on endpoint)\n\nSame keyword arguments as FastAPI\'s url_path_for, except string arguments enclosed in < > will be interpreted as attributes to pull from the object. For example, here we need to pass an `item_id` argument as required by our endpoint function, and we want to populate that with our item object\'s `id` attribute.\n\n### Create a link set\n\nIn some cases we want to create a map of relational links. In these cases we can create a `LinkSet` field describing each link and it\'s relationship to the object. The `LinkSet` class is really just a spicy dictionary that tells the parent `HyperModel` to "render" each link in the link set, and includes some extra OpenAPI schema stuff.\n\n```python\nclass Person(HyperModel):\n    id: str\n    name: str\n    items: List[ItemSummary]\n\n    href = UrlFor("read_person", {"person_id": "<id>"})\n    links = LinkSet(\n        {\n            "self": UrlFor("read_person", {"person_id": "<id>"}),\n            "items": UrlFor("read_person_items", {"person_id": "<id>"}),\n        }\n    )\n```\n\n### Putting it all together\n\nFor this example, we can make a dictionary containing some fake data, and add extra models, even nesting models if we want. A complete example based on this documentation can be found [here](examples/simple_app.py).\n\nIf we run the example application and go to our `/items` URL, we should get a response like:\n\n```json\n[\n  {\n    "name": "Foo",\n    "id": "item01",\n    "href": "/items/item01"\n  },\n  {\n    "name": "Bar",\n    "id": "item02",\n    "href": "/items/item02"\n  },\n  {\n    "name": "Baz",\n    "id": "item03",\n    "href": "/items/item03"\n  }\n]\n```\n\n## Limitations\n\nCurrently, query parameters will not resolve correctly. When generating a resource URL, ensure all parameters passed are path parameters, not query parameters.\n\nThis is an upstream issue, being tracked [here](https://github.com/encode/starlette/issues/560).\n\n## Attributions\n\nSome functionality is based on [Flask-Marshmallow](https://github.com/marshmallow-code/flask-marshmallow/blob/dev/src/flask_marshmallow/fields.py) `URLFor` class.\n',
    'author': 'Joel Collins',
    'author_email': 'joel.collins@renalregistry.nhs.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jtc42/fastapi-hypermodel',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<4.0',
}


setup(**setup_kwargs)
